from pkg_resources import DistributionNotFound, get_distribution

from .geod import direct, inverse


try:
    __version__ = get_distribution('GeoVectors').version
except DistributionNotFound:
    __version__ = '(local)'

__all__ = ['direct', 'inverse']



# pylint: disable=redefined-outer-name,unused-variable,expression-not-assigned,singleton-comparison
import numpy as np
from geographiclib.geodesic import Geodesic

from geovectorslib import geod


def describe_inverse():
    def fixed(expect):
        r = geod.inverse(
            [-90, -89, 0, 89, 90, 0, 0],
            [-180, -170, 0, 150, 180, 0, -180],
            [90, 89, 0, -89, -90, 0, 0],
            [180, 170, 0, -150, -180, -180, 180],
        )
        t = {
            's12': [
                20003931.4586233,
                19783938.13763718,
                0.0,
                19810477.03907933,
                20003931.4586233,
                20003931.4586233,
                0.0,
            ],
            'azi1': [0.0, 349.99759033, np.nan, 149.99358041, 0.0, 0.0, np.nan],
            'azi2': [180.0, 349.99759033, np.nan, 149.99358041, 180.0, 180.0, np.nan],
            'iterations': 3,
        }
        np.testing.assert_allclose(r['s12'], t['s12'], rtol=1e-9, atol=1e-9)
        np.testing.assert_allclose(r['azi1'], t['azi1'], rtol=1e-9, atol=1e-9)
        np.testing.assert_allclose(r['azi2'], t['azi2'], rtol=1e-9, atol=1e-9)
        expect(r['iterations']) == t['iterations']

    def vs_geographiclib():
        glib = Geodesic.WGS84

        lats = np.arange(-90, 90, 45)
        lons = np.arange(-180, 180, 30)
        vp1 = [(p1, p2) for p1 in lats for p2 in lons]
        vp2 = [(1, 180) for _ in range(len(vp1))]

        g1s = [glib.Inverse(*vp1[i], *vp2[i]) for i in range(len(vp1))]
        g2s = geod.inverse(*list(zip(*vp1)), *list(zip(*vp2)))

        s12 = [g1s[i]['s12'] for i in range(len(g1s))]

        # not testing azimuth, as it varies from 0-360deg.
        np.testing.assert_allclose(g2s['s12'], s12, rtol=1e-5, atol=1e-5)

    def large_size():
        np.random.seed(42)
        lats1 = np.random.uniform(-90, 90, 100000)
        lons1 = np.random.uniform(-180, 180, 100000)
        lats2 = np.random.uniform(-90, 90, 100000)
        lons2 = np.random.uniform(-180, 180, 100000)
        _ = geod.inverse(lats1, lons1, lats2, lons2)

    def near_antipodal():
        glib = Geodesic.WGS84

        lats1 = np.array([0, 0, 16.24568372, 0])
        lons1 = np.array([0, -180, 124.84613035, 0])
        lats2 = np.array([0, 0, -16.70728358, 0.5])
        lons2 = np.array([180, 180, -55.2234313, 179.7])

        vInverse = np.vectorize(glib.Inverse)

        g1s = vInverse(lats1, lons1, lats2, lons2)
        g2s = geod.inverse(lats1, lons1, lats2, lons2)

        s12 = [g1s[i]['s12'] for i in range(len(g1s))]
        np.testing.assert_allclose(g2s['s12'], s12, rtol=1e-10, atol=1e-10)


def describe_direct():
    def calculate(expect):
        r = geod.direct(
            [-90, -89, 0, 89, 90],
            [-180, -170, 0, 150, 180],
            [40, 50, 300, 360, 30],
            [5000.0, 3.0e6, 6.0e6, 10.0e6, 20.0e6],
        )
        t = {
            'lat2': [-89.95523483, -62.46780606, 23.95378158, 1.02790157, -89.96480152],
            'lon2': [-140.0, -121.47529512, -49.91980895, -30.0, -30.0],
            'azi2': [
                2.88637944e-12,
                1.65855758e00,
                2.88716163e02,
                1.80000000e02,
                1.80000000e02,
            ],
            'iterations': 4,
        }
        np.testing.assert_allclose(r['lat2'], t['lat2'], rtol=1e-8, atol=1e-8)
        np.testing.assert_allclose(r['lon2'], t['lon2'], rtol=1e-8, atol=1e-8)
        np.testing.assert_allclose(r['azi2'], t['azi2'], rtol=1e-8, atol=1e-8)

        expect(r['iterations']) == t['iterations']

    def vs_geographiclib():
        glib = Geodesic.WGS84

        npoints = 20
        lats = np.arange(-90, 90, 180 / npoints)
        lons = np.arange(-180, 180, 360 / npoints)
        dist = np.arange(0, 6000.0e3, 6000.0e3 / npoints)
        brgs = np.arange(-20, 360, 380 / npoints)

        g1s = [glib.Direct(lats[i], lons[i], brgs[i], dist[i]) for i in range(npoints)]
        g2s = geod.direct(lats, lons, brgs, dist)

        lats2 = np.array([g1s[i]['lat2'] for i in range(len(g1s))])

        # not testing lons as the -180;180 range is not consistent
        np.testing.assert_allclose(g2s['lat2'], lats2)

    def large_size():
        np.random.seed(42)
        lats1 = np.random.uniform(-90, 90, 100000)
        lons1 = np.random.uniform(-180, 180, 100000)
        brgs = np.random.uniform(0, 360, 100000)
        dist = np.random.uniform(0, 20e6, 100000)
        _ = geod.inverse(lats1, lons1, brgs, dist)



import numpy as np


def wrap90deg(deg: 'np.ndarray') -> 'np.ndarray':
    """
    Make sure that degrees (e.g. lats) is within -90;90 range.

    Ref: https://www.movable-type.co.uk/scripts/latlong-vincenty.html
    """
    mask = (deg < -90) | (deg > 90)
    deg[mask] = np.abs((deg[mask] % 360 + 270) % 360 - 180) - 90
    return deg


def wrap180deg(deg: 'np.ndarray') -> 'np.ndarray':
    """
    Make sure that degrees (e.g. lon) is within +/-(0-180) range.

    Ref: https://www.movable-type.co.uk/scripts/latlong-vincenty.html
    """
    mask = (deg <= -180) | (deg >= 180)
    deg[mask] = (deg[mask] + 540) % 360 - 180
    return deg


def wrap360deg(deg: 'np.ndarray') -> 'np.ndarray':
    """
    Make sure that degrees (e.g. heading) is within 360 range.

    Ref: https://www.movable-type.co.uk/scripts/latlong-vincenty.html
    """
    mask = (deg < 0) | (deg > 360)
    deg[mask] = (deg[mask] % 360 + 360) % 360
    return deg



import numpy as np
from geographiclib.geodesic import Geodesic

from geovectorslib.utils import wrap90deg, wrap180deg, wrap360deg


def inverse(lats1: 'list', lons1: 'list', lats2: 'list', lons2: 'list') -> 'dict':
    """
    Inverse geodesic using Vincenty approach.

    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees)

    All args must be of equal length.

    Ref:
    https://www.movable-type.co.uk/scripts/latlong-vincenty.html
    """

    eps = np.finfo(float).eps

    lon1, lon2 = map(wrap180deg, [np.array(lons1), np.array(lons2)])
    lat1, lat2 = map(wrap90deg, [np.array(lats1), np.array(lats2)])
    lon1, lat1, lon2, lat2 = map(np.radians, [lon1, lat1, lon2, lat2])

    # WGS84
    a = 6378137.0
    b = 6356752.314245
    f = 1.0 / 298.257223563

    # L = difference in longitude
    L = lon2 - lon1

    # U = reduced latitude, defined by tan U = (1-f)·tanφ.
    tan_U1 = (1 - f) * np.tan(lat1)
    cos_U1 = 1 / np.sqrt((1 + tan_U1 * tan_U1))
    sin_U1 = tan_U1 * cos_U1
    tan_U2 = (1 - f) * np.tan(lat2)
    cos_U2 = 1 / np.sqrt((1 + tan_U2 * tan_U2))
    sin_U2 = tan_U2 * cos_U2

    # checks for antipodal points
    antipodal = (np.abs(L) > np.pi / 2) | (np.abs(lat2 - lat1) > np.pi / 2)

    # delta = difference in longitude on an auxiliary sphere
    delta = L
    sin_delta = np.sin(delta)
    cos_delta = np.cos(delta)

    # sigma = angular distance P₁ P₂ on the sphere
    sigma = np.zeros(lat1.shape)
    sigma[antipodal] = np.pi

    cos_sigma = np.ones(lat1.shape)
    cos_sigma[antipodal] = -1

    # sigmam = angular distance on the sphere from the equator
    # to the midpoint of the line
    # azi = azimuth of the geodesic at the equator
    delta_prime = np.zeros(lat1.shape)
    iterations = 0

    # init before loop to allow mask indexing
    sin_Sq_sigma = np.zeros(lat1.shape)
    sin_sigma = np.zeros(lat1.shape)
    cos_sigma = np.zeros(lat1.shape)
    sin_azi = np.zeros(lat1.shape)
    cos_Sq_azi = np.ones(lat1.shape)
    cos_2_sigma_m = np.ones(lat1.shape)
    C = np.ones(lat1.shape)

    # init mask
    m = np.ones(lat1.shape, dtype=bool)
    conv_mask = np.zeros(lat1.shape, dtype=bool)

    while (np.abs(delta[m] - delta_prime[m]) > 1e-12).any():
        sin_delta = np.sin(delta)
        cos_delta = np.cos(delta)
        sin_Sq_sigma = (cos_U2 * sin_delta) * (cos_U2 * sin_delta) + (
            cos_U1 * sin_U2 - sin_U1 * cos_U2 * cos_delta
        ) * (cos_U1 * sin_U2 - sin_U1 * cos_U2 * cos_delta)

        # co-incident/antipodal points mask - exclude from the rest of the loop
        # the value has to be about 1.e-4 experimentally
        # otherwise there are issues with convergence for near antipodal points
        m = (np.abs(sin_Sq_sigma) > eps) & (sin_Sq_sigma != np.nan)

        sin_sigma[m] = np.sqrt(sin_Sq_sigma[m])
        cos_sigma[m] = sin_U1[m] * sin_U2[m] + cos_U1[m] * cos_U2[m] * cos_delta[m]
        sigma[m] = np.arctan2(sin_sigma[m], cos_sigma[m])
        sin_azi[m] = cos_U1[m] * cos_U2[m] * sin_delta[m] / sin_sigma[m]
        cos_Sq_azi[m] = 1 - sin_azi[m] * sin_azi[m]

        # on equatorial line cos²azi = 0
        cos_2_sigma_m[m] = cos_sigma[m] - 2 * sin_U1[m] * sin_U2[m] / cos_Sq_azi[m]
        cos_2_sigma_m[cos_Sq_azi == 0] = 0

        C[m] = f / 16 * cos_Sq_azi[m] * (4 + f * (4 - 3 * cos_Sq_azi[m]))

        delta_prime = delta
        delta = L + (1 - C) * f * sin_azi * (
            sigma
            + (C * sin_sigma)
            * (cos_2_sigma_m + C * cos_sigma * (-1 + 2 * cos_2_sigma_m * cos_2_sigma_m))
        )

        # Exceptions
        iteration_check = np.abs(delta)
        iteration_check[antipodal] = iteration_check[antipodal] - np.pi
        if (iteration_check > np.pi).any():
            raise Exception('delta > np.pi')

        iterations += 1
        if iterations >= 50:
            conv_mask = np.abs(delta - delta_prime) > 1e-12
            break

    uSq = cos_Sq_azi * (a * a - b * b) / (b * b)
    A = 1 + uSq / 16384 * (4096 + uSq * (-768 + uSq * (320 - 175 * uSq)))
    B = uSq / 1024 * (256 + uSq * (-128 + uSq * (74 - 47 * uSq)))
    d_sigma = (B * sin_sigma) * (
        cos_2_sigma_m
        + (B / 4)
        * (
            cos_sigma * (-1 + 2 * cos_2_sigma_m * cos_2_sigma_m)
            - (B / 6 * cos_2_sigma_m)
            * (-3 + 4 * sin_sigma * sin_sigma)
            * (-3 + 4 * cos_2_sigma_m * cos_2_sigma_m)
        )
    )
    # length of the geodesic
    s12 = b * A * (sigma - d_sigma)

    # note special handling of exactly antipodal points where sin²sigma = 0
    # (due to discontinuity atan2(0, 0) = 0 but atan2(eps, 0) = np.pi/2 / 90°)
    # in which case bearing is always meridional, due north (or due south!)

    azi1 = np.arctan2(cos_U2 * sin_delta, cos_U1 * sin_U2 - sin_U1 * cos_U2 * cos_delta)
    azi1[np.abs(sin_Sq_sigma) < eps] = 0

    azi2 = np.arctan2(
        cos_U1 * sin_delta, -sin_U1 * cos_U2 + cos_U1 * sin_U2 * cos_delta
    )
    azi2[np.abs(sin_Sq_sigma) < eps] = np.pi

    azi1 = wrap360deg(azi1 * 180 / np.pi)
    azi2 = wrap360deg(azi2 * 180 / np.pi)

    # if distance is too small
    mask = s12 < eps
    azi1[mask] = None
    azi2[mask] = None

    # use geographiclib for points which didn't converge
    if conv_mask[conv_mask].shape[0] > 0:
        vInverse = np.vectorize(Geodesic.WGS84.Inverse)
        _ps = vInverse(
            lats1[conv_mask], lons1[conv_mask], lats2[conv_mask], lons2[conv_mask]
        )
        s12[conv_mask] = [_p['s12'] for _p in _ps]
        azi1[conv_mask] = [_p['azi1'] for _p in _ps]
        azi2[conv_mask] = [_p['azi2'] for _p in _ps]

    return {'s12': s12, 'azi1': azi1, 'azi2': azi2, 'iterations': iterations}


def direct(lats1: 'list', lons1: 'list', brgs: 'list', dists: 'list') -> 'dict':
    """
    Direct geodesic using Vincenty approach.

    Given a start point, initial bearing (0° is North, clockwise)
    and distance in meters, this will calculate the destination point
    and final bearing travelling along a (shortest distance) great circle arc.

    Bearing in 0-360deg starting from North clockwise.

    all variables must be of same length

    Ref:
    https://www.movable-type.co.uk/scripts/latlong.html
    https://www.movable-type.co.uk/scripts/latlong-vincenty.html
    """

    lon1, lat1, brg = map(np.radians, [lons1, lats1, brgs])

    # WGS84
    a = 6378137.0
    b = 6356752.314245
    f = 1.0 / 298.257223563

    azi1 = brg
    s = dists

    sin_azi1 = np.sin(azi1)
    cos_azi1 = np.cos(azi1)

    tan_U1 = (1 - f) * np.tan(lat1)
    cos_U1 = 1 / np.sqrt((1 + tan_U1 * tan_U1))
    sin_U1 = tan_U1 * cos_U1

    # sigma1 = angular distance on the sphere from the equator to P1
    sigma1 = np.arctan2(tan_U1, cos_azi1)

    sin_azi = cos_U1 * sin_azi1  # azi = azimuth of the geodesic at the equator
    cos_Sq_azi = 1 - sin_azi * sin_azi
    uSq = cos_Sq_azi * (a * a - b * b) / (b * b)
    A = 1 + uSq / 16384 * (4096 + uSq * (-768 + uSq * (320 - 175 * uSq)))
    B = uSq / 1024 * (256 + uSq * (-128 + uSq * (74 - 47 * uSq)))

    sigma = s / (b * A)
    # sigma = angular distance P₁ P₂ on the sphere
    # sigmam = angular distance on the sphere from the equator to
    # the midpoint of the line
    sin_sigma = np.sin(sigma)
    cos_sigma = np.cos(sigma)
    cos_2_sigma_m = np.cos(2 * sigma1 + sigma)

    iterations = 0
    sigma_prime = 0

    while (np.abs(sigma - sigma_prime) > 1e-12).any():
        cos_2_sigma_m = np.cos(2 * sigma1 + sigma)
        sin_sigma = np.sin(sigma)
        cos_sigma = np.cos(sigma)
        delta_sigma = (B * sin_sigma) * (
            cos_2_sigma_m
            + B
            / 4
            * (
                cos_sigma * (-1 + 2 * cos_2_sigma_m * cos_2_sigma_m)
                - (B / 6 * cos_2_sigma_m)
                * (-3 + 4 * sin_sigma * sin_sigma)
                * (-3 + 4 * cos_2_sigma_m * cos_2_sigma_m)
            )
        )
        sigma_prime = sigma
        sigma = s / (b * A) + delta_sigma
        iterations += 1
        if iterations >= 50:
            raise Exception('Vincenty formula failed to converge')

    x = sin_U1 * sin_sigma - cos_U1 * cos_sigma * cos_azi1
    lat2 = np.arctan2(
        sin_U1 * cos_sigma + cos_U1 * sin_sigma * cos_azi1,
        (1 - f) * np.sqrt(sin_azi * sin_azi + x * x),
    )
    lambda_ = np.arctan2(
        sin_sigma * sin_azi1, cos_U1 * cos_sigma - sin_U1 * sin_sigma * cos_azi1
    )
    C = f / 16 * cos_Sq_azi * (4 + f * (4 - 3 * cos_Sq_azi))
    L = lambda_ - (1 - C) * f * sin_azi * (
        sigma
        + C
        * sin_sigma
        * (cos_2_sigma_m + C * cos_sigma * (-1 + 2 * cos_2_sigma_m * cos_2_sigma_m))
    )
    lon2 = lon1 + L
    azi2 = np.arctan2(sin_azi, -x)

    lat2 = wrap90deg(lat2 * 180.0 / np.pi)
    lon2 = wrap180deg(lon2 * 180.0 / np.pi)
    azi2 = wrap360deg(azi2 * 180.0 / np.pi)

    return {'lat2': lat2, 'lon2': lon2, 'azi2': azi2, 'iterations': iterations}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['geovectorslib', 'geovectorslib.tests']

package_data = \
{'': ['*']}

install_requires = \
['geographiclib>=1.50,<2.0', 'numpy>=1.18.0,<2.0.0']

entry_points = \
{'console_scripts': ['GeoVectors = geovectorslib.cli:main']}

setup_kwargs = {
    'name': 'geovectorslib',
    'version': '1.0',
    'description': 'Vectorized geodesic calculations.',
    'long_description': "# Overview\n\n[![Unix Build Status](https://img.shields.io/travis/omdv/geovectors/master.svg?label=unix)](https://travis-ci.org/omdv/geovectors)\n[![Coverage Status](https://img.shields.io/coveralls/omdv/geovectors/master.svg)](https://coveralls.io/r/omdv/geovectors)\n[![Scrutinizer Code Quality](https://img.shields.io/scrutinizer/g/omdv/geovectors.svg)](https://scrutinizer-ci.com/g/omdv/geovectors/?branch=master)\n[![PyPI Version](https://img.shields.io/pypi/v/GeoVectors.svg)](https://pypi.org/project/GeoVectors)\n[![PyPI License](https://img.shields.io/pypi/l/GeoVectors.svg)](https://pypi.org/project/GeoVectors)\n\nThis library provides vectorized direct and inverse geodesic methods.\n\nThe motivation was to have the accurate and fast vectorized geodesic routines for sailboat routing project. There are few python libraries, with [geographiclib](https://geographiclib.sourceforge.io/html/python/index.html) being the most accurate and reliable. Haversine method, which is widely used as an example of fast inverse method can be vectorized rather easily, however the errors are expected to be at [least 0.5%](https://en.wikipedia.org/wiki/Haversine_formula#Formulation). There are no vectorized AND accurate options.\n\nThis library is based on `numpy` and uses [Vincenty's formulae](https://en.wikipedia.org/wiki/Vincenty's_formulae). It is heavily based on the [Movable Type Scripts blog](https://www.movable-type.co.uk/scripts/latlong-vincenty.html) and Javascript [Geodesy](https://www.npmjs.com/package/geodesy) code.\n\nVincenty's inverse algorithm is accurate, but sensitive to [nearly antipodal points](https://en.wikipedia.org/wiki/Vincenty%27s_formulae#Nearly_antipodal_points). One approach would be to return `NaN` for such points, with the assumption that they are not frequently observed in practical applications, however as [this discussion](https://gis.stackexchange.com/questions/84885/difference-between-vincenty-and-great-circle-distance-calculations) nicely pointed out the package cannot be complete if it cannot handle these situations. I found that the issue can be solved by relaxing one of convergence criteria, but it results in errors up to 0.25% vs geographiclib for these points.\n\nSo, instead, this library uses the vectorized Vincenty's formulae with **geographiclib as a fallback** for unconverged points.\n\nSee [notebook](https://github.com/omdv/geovectors/blob/master/notebooks/demo.ipynb) for execution time comparisons vs geographiclib.\n\n```\nDirect method for 100,000 points\n\n94.9 ms ± 25 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)\nvs\n9.79 s ± 1.4 s per loop (mean ± std. dev. of 7 runs, 1 loop each)\n```\n\n```\nInverse method for 100,000 points\n\n1.5 s ± 504 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)\nvs\n24.2 s ± 3.91 s per loop (mean ± std. dev. of 7 runs, 1 loop each)\n```\n\n# Setup\n\n## Requirements\n\n* Python 3.7+\n* Numpy\n* Geographiclib\n\n## Installation\n\nInstall it directly into an activated virtual environment:\n\n```text\n$ pip install GeoVectors\n```\n\n# Usage\n\nAfter installation, the package can imported:\n\n```text\n$ python\n>>> from geovectorslib import direct, inverse\n>>> direct(lats1, lon1, bearings, distances)\n>>> inverse(lats1, lons1, lats2, lons2)\n```\n\n```text\nLatitudes in decimal degrees [-90; +90].\nLongitudes in decimal degrees [-180; +180].\nBearings in decimal degrees [0; 360].\nDistances in meters.\n```\n\n# References\n\n[Movable Type Scripts](https://www.movable-type.co.uk/scripts/latlong-vincenty.html)\n\n[Geodesy](https://www.npmjs.com/package/geodesy)\n\n[Geopy](https://pypi.org/project/geopy/)\n\n[Geographiclib](https://geographiclib.sourceforge.io/html/python/index.html)\n\n[Stackoverflow discussion](https://gis.stackexchange.com/questions/84885/difference-between-vincenty-and-great-circle-distance-calculations)",
    'author': 'Oleg Medvedev',
    'author_email': 'omdv.public@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/geovectorslib',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
